import asyncio
import functools
import operator
import aiohttp
from multidict import MultiDict
from urllib.parse import urljoin
from .base import validate, Client, Graph, Proxy, Remote, Resource


class TokenAuth(aiohttp.BasicAuth):
    __repr__ = encode = functools.partialmethod(' '.join)

    def __new__(cls, auth):
        return tuple.__new__(cls, *auth.items())


class AsyncClient(aiohttp.ClientSession):
    """An asynchronous ClientSession which sends requests to a base url.

    :param url: base url for requests
    :param trailing: trailing chars (e.g. /) appended to the url
    :param params: default query params
    :param auth: additional authorization support for ``{token_type: access_token}``,
        available per request as well
    :param attrs: additional ClientSession options, e.g., loop
    """
    __truediv__ = Client.__truediv__
    __repr__ = Client.__repr__
    get = Client.get
    options = Client.options
    head = Client.head
    post = Client.post
    put = Client.put
    patch = Client.patch
    delete = Client.delete
    headers = property(operator.attrgetter('_default_headers'), doc="default headers")
    auth = property(operator.attrgetter('_default_auth'), doc="default auth")

    def __init__(self, url, *, trailing='', params=(), auth=None, **attrs):
        if {'connector', 'loop'}.isdisjoint(attrs):
            attrs['loop'] = asyncio.get_event_loop()
        attrs['auth'] = TokenAuth(auth) if isinstance(auth, dict) else auth
        super().__init__(**attrs)
        self._attrs = attrs
        self.params = MultiDict(params)
        self.trailing = trailing
        self.url = url.rstrip('/') + '/'

    def __del__(self):  # avoids warning and race condition in parent
        if self._connector_owner:  # pragma: no branch
            self._connector.close()

    @classmethod
    def clone(cls, other, path='', **kwargs):
        url = urljoin(other.url, path)
        kwargs.update(other._attrs)
        return cls(url, trailing=other.trailing, params=other.params, **kwargs)

    def request(self, method, path, params=(), auth=None, **kwargs):
        kwargs['auth'] = TokenAuth(auth) if isinstance(auth, dict) else auth
        params = MultiDict(params)
        params.extend(self.params)
        url = urljoin(self.url, path).rstrip('/') + self.trailing
        return super().request(method, url, params=params, **kwargs)
    request.__doc__ = Client.request.__doc__

    def run(self, name, *args, **kwargs):
        """Synchronously call method and run coroutine."""
        return self.loop.run_until_complete(getattr(self, name)(*args, **kwargs))


class AsyncResource(AsyncClient):
    """An `AsyncClient`_ which returns json content and has syntactic support for requests."""
    client = property(AsyncClient.clone, doc="upcasted `AsyncClient`_")
    __getattr__ = AsyncClient.__truediv__
    __getitem__ = AsyncClient.get
    content_type = Resource.content_type
    __call__ = Resource.__call__

    def __init__(self, url, **kwargs):
        super().__init__(url, **kwargs)
        self._raise_for_status = True

    async def request(self, method, path, **kwargs):
        response = await super().request(method, path, **kwargs)
        content_type = self.content_type(response)
        if content_type == 'json':
            return await response.json(content_type='')
        return await getattr(response, content_type or 'read')()
    request.__doc__ = Resource.request.__doc__

    async def update(self, path='', callback=None, **json):
        if callback is None:
            return await self.request('PATCH', path, json=json)
        response = await super().request('GET', path)
        json = callback(await response.json(content_type=''), **json)
        return await self.request('PUT', path, json=json, headers=validate(response))
    update.__doc__ = Resource.update.__doc__

    async def authorize(self, path='', **kwargs):
        method = 'GET' if {'json', 'data'}.isdisjoint(kwargs) else 'POST'
        result = await self.request(method, path, **kwargs)
        self._default_auth = TokenAuth({result['token_type']: result['access_token']})
        return result
    authorize.__doc__ = Resource.authorize.__doc__


class AsyncRemote(AsyncClient):
    """An `AsyncClient`_ which defaults to posts with json bodies, i.e., RPC.

    :param url: base url for requests
    :param json: default json body for all calls
    :param kwargs: same options as `AsyncClient`_
    """
    client = AsyncResource.client
    __getattr__ = AsyncResource.__getattr__
    check = staticmethod(Remote.check)

    def __init__(self, url, json=(), **kwargs):
        super().__init__(url, **kwargs)
        self._raise_for_status = True
        self.json = dict(json)

    @classmethod
    def clone(cls, other, path=''):
        return AsyncClient.clone.__func__(cls, other, path, json=other.json)

    async def __call__(self, path='', **json):
        """POST request with json body and check result."""
        response = await self.post(path, json=dict(self.json, **json))
        return self.check(await response.json(content_type=''))


class AsyncGraph(AsyncRemote):
    """An `AsyncRemote`_ client which executes GraphQL queries."""
    Error = aiohttp.ClientPayloadError
    check = classmethod(Graph.check.__func__)
    execute = Graph.execute


class AsyncProxy(AsyncClient):
    """An extensible embedded proxy client to multiple hosts.

    The default implementation provides load balancing based on active connections.
    It does not provide error handling or retrying.

    :param urls: base urls for requests
    :param kwargs: same options as `AsyncClient`_
    """
    Stats = Proxy.Stats
    priority = Proxy.priority
    choice = Proxy.choice

    def __init__(self, *urls, **kwargs):
        super().__init__('', **kwargs)
        self.urls = {(url.rstrip('/') + '/'): self.Stats() for url in urls}

    @classmethod
    def clone(cls, other, path=''):
        urls = (urljoin(url, path) for url in other.urls)
        return cls(*urls, trailing=other.trailing, params=other.params, **other._attrs)

    async def request(self, method, path, **kwargs):
        url = self.choice(method)
        with self.urls[url] as stats:
            response = await super().request(method, urljoin(url, path), **kwargs)
        stats.add(failures=int(response.status >= 500))
        return response
    request.__doc__ = Proxy.request.__doc__



from .base import Client, Graph, Proxy, Remote, Resource  # noqa
try:
    from .aio import AsyncClient, AsyncGraph, AsyncProxy, AsyncRemote, AsyncResource  # noqa
except SyntaxError:  # pragma: no cover
    pass

__version__ = '1.0'


def singleton(*args, **kwargs):
    """Return a decorator for singleton class instances."""
    return lambda cls: cls(*args, **kwargs)



import pytest
import requests
import clients

aio = getattr(clients, 'aio', None)
pytest_plugins = 'httpbin',


def pytest_ignore_collect(path, config):
    return not aio and path.basename == 'test_aio.py'


def pytest_report_header(config):
    modules = {requests, getattr(aio, 'aiohttp', requests)}
    return ', '.join('{} {}'.format(module.__name__, module.__version__) for module in modules)


@pytest.fixture
def url(httpbin):
    return httpbin.url



import asyncio
import json
import operator
import aiohttp
import pytest
import clients


def results(*coros):
    loop = asyncio.get_event_loop()
    fs = [asyncio.ensure_future(coro, loop=loop) for coro in coros]
    return map(loop.run_until_complete, fs)


def test_client():
    client = clients.AsyncClient('http://httpbin.org/', params={'q': 0})
    coros = (client.head(), client.options(), client.post('post'), client.put('put'),
             client.patch('patch'), client.delete('delete'), (client / 'ip').get(params={'q': 1}))
    for r in results(*coros):
        assert r.status == 200 and r.url.query_string.endswith('q=0')
    assert r.url.query_string == 'q=1&q=0'
    data, = results(r.json())
    assert set(data) == {'origin'}


def test_resource():
    params = {'etag': 'W/0', 'last-modified': 'now'}
    resource = clients.AsyncResource('http://httpbin.org/', params=params)
    it = results(resource['robots.txt'], resource.bytes('1'),
                 resource.update('patch', key='value'), resource.status('404'),
                 resource.update('response-headers', callback=dict, key='value'))
    assert isinstance(next(it), str)
    assert isinstance(next(it), bytes)
    assert next(it)['json'] == {'key': 'value'}
    with pytest.raises(aiohttp.ClientError, match='404'):
        next(it)
    with pytest.raises(aiohttp.ClientError, match='405'):
        next(it)


def test_content(url):
    resource = clients.AsyncResource(url)
    resource.content_type = lambda response: 'json'
    coro = resource.get('robots.txt')
    assert not hasattr(coro, '__aenter__')
    with pytest.raises(ValueError):
        data, = results(coro)


def test_authorize(url, monkeypatch):
    resource = clients.AsyncResource(url, auth=aiohttp.BasicAuth('', ''))
    basic, token = results(resource.get('headers'), resource.get('headers', auth={'token': 'abc123'}))
    assert basic['headers']['Authorization'].startswith('Basic ')
    assert token['headers']['Authorization'] == 'token abc123'
    resource = clients.AsyncResource(url, auth={'token': 'abc123'})
    assert resource.headers == {} and resource.auth == ('token', 'abc123')
    basic, token = results(resource.get('headers', auth=aiohttp.BasicAuth('', '')), resource.get('headers'))
    assert basic['headers']['Authorization'].startswith('Basic ')
    assert token['headers']['Authorization'] == 'token abc123'

    resource = clients.AsyncResource(url)
    future = asyncio.Future()
    future.set_result({'access_token': 'abc123', 'token_type': 'Bearer', 'expires_in': 0})
    monkeypatch.setattr(clients.AsyncResource, 'request', lambda *args, **kwargs: future)
    for key in ('params', 'data', 'json'):
        assert resource.run('authorize', **{key: {}}) == future.result()
        assert resource.auth == ('Bearer', 'abc123')


def test_remote(url):
    remote = clients.AsyncRemote(url, json={'key': 'value'})
    result, = results(remote('post'))
    assert result['json'] == {'key': 'value'}
    clients.AsyncRemote.check = operator.methodcaller('pop', 'json')
    result, = results((remote / 'post')(name='value'))
    assert result == {'key': 'value', 'name': 'value'}


def test_graph(url):
    graph = clients.AsyncGraph(url).anything
    data, = results(graph.execute('{ viewer { login }}'))
    assert json.loads(data) == {'query': '{ viewer { login }}', 'variables': {}}
    with pytest.raises(aiohttp.ClientPayloadError, match='reason'):
        clients.AsyncGraph.check({'errors': ['reason']})


def test_proxy(url):
    proxy = clients.AsyncProxy(url, 'http://httpbin.org/')
    responses = results(*(proxy.get('get') for _ in proxy.urls))
    urls = {response.url: response.json() for response in responses}
    assert len(urls) == len(proxy.urls)
    assert all(results(*urls.values()))

    fs = (proxy.get('status/500') for _ in proxy.urls)
    response, = results(next(fs))
    assert next(results(*fs)).url != response.url

    proxy = clients.AsyncProxy('http://localhost/', 'http://httpbin.org/')
    with pytest.raises(aiohttp.ClientError):
        list(results(*(proxy.get() for _ in proxy.urls)))
    response, = results(proxy.get())
    assert response.status == 200


def test_clones():
    client = clients.AsyncClient('http://localhost/', trailing='/')
    assert str(client) == 'AsyncClient(http://localhost/... /)'
    assert str(client / 'path') == 'AsyncClient(http://localhost/path/... /)'

    resource = clients.AsyncResource('http://localhost/').path
    assert str(resource) == 'AsyncResource(http://localhost/path/... )'
    assert type(resource.client) is clients.AsyncClient

    remote = clients.AsyncRemote('http://localhost/').path
    assert str(remote) == 'AsyncRemote(http://localhost/path/... )'
    assert type(remote.client) is clients.AsyncClient

    proxy = clients.AsyncProxy('http://localhost/', 'http://127.0.0.1') / 'path'
    assert str(proxy) == 'AsyncProxy(/... )'



import collections
import functools
import random
import re
import threading
import requests
from requests.compat import json, urljoin
try:
    from future_builtins import map
except ImportError:  # pragma: no cover
    pass


def content_type(response, **patterns):
    """Return name for response's content-type based on regular expression matches."""
    ct = response.headers.get('content-type', '')
    matches = (name for name, pattern in patterns.items() if re.match(pattern, ct))
    return next(matches, '')


def validate(response):
    """Return validation headers from response translated for modification."""
    headers = response.headers
    validators = {'etag': 'if-match', 'last-modified': 'if-unmodified-since'}
    return {validators[key]: headers[key] for key in validators if key in headers}


class TokenAuth(dict):
    def __call__(self, req):
        req.headers['Authorization'] = ' '.join(*self.items())
        return req


class Client(requests.Session):
    """A Session which sends requests to a base url.

    :param url: base url for requests
    :param trailing: trailing chars (e.g. /) appended to the url
    :param headers: additional headers to include in requests
    :param auth: additional authorization support for ``{token_type: access_token}``,
        available per request as well
    :param attrs: additional Session attributes
    """
    def __init__(self, url, trailing='', headers=(), auth=None, **attrs):
        super(Client, self).__init__()
        self.__setstate__(attrs)
        self.auth = TokenAuth(auth) if isinstance(auth, dict) else auth
        self.headers.update(headers)
        self.trailing = trailing
        self.url = url.rstrip('/') + '/'

    @classmethod
    def clone(cls, other, path='', **kwargs):
        kwargs.update(other.__getstate__())
        return cls(urljoin(other.url, path), trailing=other.trailing, **kwargs)

    def __repr__(self):
        return '{}({}... {})'.format(type(self).__name__, self.url, self.trailing)

    def __truediv__(self, path):
        """Return a cloned client with appended path."""
        return type(self).clone(self, path)
    __div__ = __truediv__

    def request(self, method, path, auth=None, **kwargs):
        """Send request with relative or absolute path and return response."""
        kwargs['auth'] = TokenAuth(auth) if isinstance(auth, dict) else auth
        url = urljoin(self.url, path).rstrip('/') + self.trailing
        return super(Client, self).request(method, url, **kwargs)

    def get(self, path='', **kwargs):
        """GET request with optional path."""
        return self.request('GET', path, **kwargs)

    def options(self, path='', **kwargs):
        """OPTIONS request with optional path."""
        return self.request('OPTIONS', path, **kwargs)

    def head(self, path='', allow_redirects=False, **kwargs):
        """HEAD request with optional path."""
        return self.request('HEAD', path, allow_redirects=allow_redirects, **kwargs)

    def post(self, path='', json=None, **kwargs):
        """POST request with optional path and json body."""
        return self.request('POST', path, json=json, **kwargs)

    def put(self, path='', json=None, **kwargs):
        """PUT request with optional path and json body."""
        return self.request('PUT', path, json=json, **kwargs)

    def patch(self, path='', json=None, **kwargs):
        """PATCH request with optional path and json body."""
        return self.request('PATCH', path, json=json, **kwargs)

    def delete(self, path='', **kwargs):
        """DELETE request with optional path."""
        return self.request('DELETE', path, **kwargs)


class Resource(Client):
    """A `Client`_ which returns json content and has syntactic support for requests."""
    client = property(Client.clone, doc="upcasted `Client`_")
    __getitem__ = Client.get
    __setitem__ = Client.put
    __delitem__ = Client.delete
    content_type = functools.partial(content_type, text='text/', json=r'application/(\w|\.)*\+?json')

    def __getattr__(self, name):
        if name in type(self).__attrs__:
            raise AttributeError(name)
        return self / name
    __getattr__.__doc__ = Client.__truediv__.__doc__

    def request(self, method, path, **kwargs):
        """Send request with path and return processed content."""
        response = super(Resource, self).request(method, path, **kwargs)
        response.raise_for_status()
        if self.content_type(response) == 'json':
            return response.json()
        return response.text if response.encoding else response.content

    def iter(self, path='', **kwargs):
        """Iterate lines or chunks from streamed GET request."""
        response = super(Resource, self).request('GET', path, stream=True, **kwargs)
        response.raise_for_status()
        content_type = self.content_type(response)
        if content_type == 'json':
            response.encoding = response.encoding or 'utf8'
            return map(json.loads, response.iter_lines(decode_unicode=True))
        if response.encoding or content_type == 'text':
            return response.iter_lines(decode_unicode=response.encoding)
        return iter(response)
    __iter__ = iter

    def __contains__(self, path):
        """Return whether endpoint exists according to HEAD request."""
        return super(Resource, self).request('HEAD', path, allow_redirects=False).ok

    def __call__(self, path='', **params):
        """GET request with params."""
        return self.get(path, params=params)

    def updater(self, path='', **kwargs):
        response = super(Resource, self).request('GET', path, **kwargs)
        headers = dict(kwargs.pop('headers', {}))
        while True:
            response.raise_for_status()
            kwargs['json'] = yield response.json()
            headers.update(validate(response))
            response = super(Resource, self).request('PUT', path, headers=headers, **kwargs)

    def update(self, path='', callback=None, **json):
        """PATCH request with json params.

        :param callback: optionally update with GET and validated PUT.
            ``callback`` is called on the json result with keyword params, i.e.,
            ``dict`` correctly implements the simple update case.
        """
        if callback is None:
            return self.patch(path, json=json)
        updater = self.updater(path)
        return updater.send(callback(next(updater), **json))

    def create(self, path='', json=None, **kwargs):
        """POST request and return location."""
        response = super(Resource, self).request('POST', path, json=json, **kwargs)
        response.raise_for_status()
        return response.headers.get('location')

    def download(self, file, path='', **kwargs):
        """Output streamed GET request to file."""
        response = super(Resource, self).request('GET', path, stream=True, **kwargs)
        response.raise_for_status()
        for chunk in response:
            file.write(chunk)
        return file

    def authorize(self, path='', **kwargs):
        """Acquire oauth access token and set ``auth``."""
        method = 'GET' if {'json', 'data'}.isdisjoint(kwargs) else 'POST'
        result = self.request(method, path, **kwargs)
        self.auth = TokenAuth({result['token_type']: result['access_token']})
        return result


class Remote(Client):
    """A `Client`_ which defaults to posts with json bodies, i.e., RPC.

    :param url: base url for requests
    :param json: default json body for all calls
    :param kwargs: same options as `Client`_
    """
    client = Resource.client
    __getattr__ = Resource.__dict__['__getattr__']

    def __init__(self, url, json=(), **kwargs):
        super(Remote, self).__init__(url, **kwargs)
        self.json = dict(json)

    @classmethod
    def clone(cls, other, path=''):
        return Client.clone.__func__(cls, other, path, json=other.json)

    def __call__(self, path='', **json):
        """POST request with json body and :meth:`check` result."""
        response = self.post(path, json=dict(self.json, **json))
        response.raise_for_status()
        return self.check(response.json())

    @staticmethod
    def check(result):
        """Override to return result or raise error, for APIs which don't use status codes."""
        return result


class Graph(Remote):
    """A `Remote`_ client which executes GraphQL queries."""
    Error = requests.HTTPError

    @classmethod
    def check(cls, result):
        """Return ``data`` or raise ``errors``."""
        for error in result.get('errors', ()):
            raise cls.Error(error)
        return result.get('data')

    def execute(self, query, **variables):
        """Execute query over POST."""
        return self(query=query, variables=variables)


class Stats(collections.Counter):
    """Thread-safe Counter.

    Context manager tracks number of active connections and errors.
    """
    def __init__(self):
        self.lock = threading.Lock()

    def add(self, **kwargs):
        """Atomically add data."""
        with self.lock:
            self.update(kwargs)

    def __enter__(self):
        self.add(connections=1)
        return self

    def __exit__(self, *args):
        self.add(connections=-1, errors=int(any(args)))


class Proxy(Client):
    """An extensible embedded proxy client to multiple hosts.

    The default implementation provides load balancing based on active connections.
    It does not provide error handling or retrying.

    :param urls: base urls for requests
    :param kwargs: same options as `Client`_
    """
    Stats = Stats

    def __init__(self, *urls, **kwargs):
        super(Proxy, self).__init__('', **kwargs)
        self.urls = {(url.rstrip('/') + '/'): self.Stats() for url in urls}

    @classmethod
    def clone(cls, other, path=''):
        urls = (urljoin(url, path) for url in other.urls)
        return cls(*urls, trailing=other.trailing, **other.__getstate__())

    def priority(self, url):
        """Return comparable priority for url.

        Minimizes errors, failures (500s), and active connections.
        None may be used to eliminate from consideration.
        """
        stats = self.urls[url]
        return tuple(stats[key] for key in ('errors', 'failures', 'connections'))

    def choice(self, method):
        """Return chosen url according to priority.

        :param method: placeholder for extensions which distinguish read/write requests
        """
        priorities = collections.defaultdict(list)
        for url in self.urls:
            priorities[self.priority(url)].append(url)
        priorities.pop(None, None)
        return random.choice(priorities[min(priorities)])

    def request(self, method, path, **kwargs):
        """Send request with relative or absolute path and return response."""
        url = self.choice(method)
        with self.urls[url] as stats:
            response = super(Proxy, self).request(method, urljoin(url, path), **kwargs)
        stats.add(failures=int(response.status_code >= 500))
        return response



from setuptools import setup

setup(
    name='clients',
    version='1.0',
    description='HTTP for humanitarians.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Aric Coady',
    author_email='aric.coady@gmail.com',
    url='https://github.com/coady/clients',
    project_urls={
        'Documentation': 'https://clients.readthedocs.io',
    },
    license='Apache Software License',
    packages=['clients'],
    install_requires=['requests>=2.4.2'],
    extras_require={':python_version>="3.5"': ['aiohttp']},
    python_requires='>=2.7',
    tests_require=['pytest>=3.7.2', 'pytest-cov', 'pytest-httpbin'],
    keywords='requests sessions responses resources asyncio',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP :: Session',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
